package lsedit;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Event;
import java.awt.Rectangle;

import java.awt.event.MouseEvent;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.swing.JComponent;
import javax.swing.undo.UndoableEdit;

public class RelationInstance extends LandscapeObject 
{
	// Constants

	public    final static int    CLIENT_SUPPLIER_EL_LEN = 16;
	protected final static int	  NEAR_PIXEL_SIZE		 = 6;

	public final static String	  LABEL_ID               = "label";
	public final static String    DESC_ID                = "description";

	// Bits set in m_marks

	public    final static int STYLE_BITS           = 0x00007;
	protected final static int LOOP_MARK	    	= 0x00010;	// Repaint this edge
	protected final static int NORMAL_MARK			= 0x00020;
	protected final static int ELISION_MARK         = 0x00040;
	protected final static int OUT_ELIDED_MARK		= 0x00100;
	protected final static int IN_ELIDED_MARK		= 0x00200;
	protected final static int IN_DIAGRAM_MARK		= 0x00400;

	public final static int DIFFERENT_REAL_SRC_MARK = 0x01000;
	public final static int DIFFERENT_REAL_DST_MARK = 0x02000;
	public final static int VALIDATED_MARK			= 0x04000;
	public final static int	LIFTED_MARK			    = 0x08000;

	public final static int GROUP_FLAG_MARK         = EntityInstance.GROUP_MARK	/* 0x10000 */;
	protected final static int VALID_MARK           = 0x20000;
	public final static int HIGHLIGHT_FLAG_MARK     = 0x40000;

	public final static int PERMANENT_MARKS         = STYLE_BITS;
	public final static int IN_OUT_ELIDED           = IN_ELIDED_MARK  | OUT_ELIDED_MARK;
	public final static int VARIOUS_TYPE_MARKS      = LOOP_MARK | NORMAL_MARK | IN_OUT_ELIDED;
	public final static int PRESENTATION_MARKS      = GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK;

//	private static int		m_totalRelations        = 0;

	// An edge is defined by a relation class and two entity instances

	protected EntityInstance  m_src,     m_dst;			// The real source and destination of this edge

	// Variables

	protected int			  m_mark = Util.LINE_STYLE_UNDEFINED;		// A collection of bits
	protected int			  m_freq;

	// ---------------
	// Wrapper methods
	// ---------------

	public EntityInstance drawSrc()
	{
		return m_src.m_drawEntity;
	}

	public EntityInstance drawDst()
	{
		return m_dst.m_drawEntity;
	}

	public RelationComponent neededComponent()
	{
		RelationComponent relationComponent = (RelationComponent) getSwingObject();
		if (relationComponent == null) {
			relationComponent = new RelationComponent(this);
		}
		return(relationComponent);
	} 

	public void repaint()
	{
		JComponent relationComponent = getSwingObject();

		if (relationComponent != null) {
			relationComponent.repaint();
	}	}
		
	public void validate()
	{
		JComponent relationComponent = getSwingObject();

		if (relationComponent != null) {
			relationComponent.validate();
	}	}

	public void frameRelation(Rectangle frame, boolean clear_flag)
	{
		RelationClass		rc;

		rc = getRelationClass();

		if (rc.isClassVisible()) {
			JComponent relationComponent = getSwingObject();

			if (relationComponent != null && ((RelationComponent) relationComponent).isFramedBy(frame)) {
				if (clear_flag) {
					clearGroupFlag();
				} else {
					setGroupAndHighlightFlag();
				}
			}
		}
	}

	// First order attributes

	public boolean processFirstOrder(String id, String value) 
	{
		if (id.equals(STYLE_ID)) {
			if (value != null) {
				setStyle(Attribute.parseIntValue(value));
			}
			return true;
		}

		return super.processFirstOrder(id, value);
	}

	public static void reportFirstOrderAttributes(ResultBox resultBox)
	{
		resultBox.addText(COLOR_ID);
		resultBox.addText(STYLE_ID);
	}

	// --------------
	// Public methods
	// --------------

/*
	protected void finalize() throws Throwable
	{
		--m_totalRelations;
	}
*/

	public RelationInstance(RelationClass parentClass, EntityInstance src, EntityInstance dst)
	{
		setParentClass(parentClass);
		this.m_src       = src;
		this.m_dst       = dst;
//		++m_totalRelations;
	}

/*
 	public static int totalRelations()
	{
		return m_totalRelations;
	}
*/

	public int getStyle()
	{	
		int	ret = m_mark & STYLE_BITS;
		if (ret == Util.LINE_STYLE_UNDEFINED) {
			ret = -1;
		}
		return ret;
	}

	public void setStyle(int value)
	{
		m_mark &= ~STYLE_BITS;
		m_mark |= (value & STYLE_BITS);
	}

	public String getRelationLabel()
	{
		Attribute attribute = getLsAttribute(LABEL_ID);
		
		if (attribute != null) {
			return attribute.parseString();
		}
		return null;
	}

	public String getRelationTooltip()
	{
		String	id;

		if (ArrowDimensions.isShowEdgeLabel()) {
			id = DESC_ID;
		} else {
			id = LABEL_ID;
		}
		Attribute attribute = getLsAttribute(id);
		
		if (attribute != null) {
			return attribute.parseString();
		}
		return null;
	}

	public String getClassLabel()
	{
		return getRelationClass().getLabel();
	}

	public String getStyleName(int style) 
	{
		return Util.getLineStyleName(style);
	}

	public void clearRelationMark(int preserve_relation_marks)
	{
		m_mark    &= preserve_relation_marks;
		setSwingObject(null);		// IJD: Make this smarter
		m_freq     = 1;

//		System.out.println("RelationInstance.clearRelationMark " + this);
	}

	public void incrementFrequency(RelationInstance ri)
	{
		++m_freq;
		if (ri.m_src != m_src) {
			orMark(DIFFERENT_REAL_SRC_MARK);
		}
		if (ri.m_dst != m_dst) {
			orMark(DIFFERENT_REAL_DST_MARK);
	}	}

	public int getFrequency()
	{
		return m_freq;
	}

	public void orMark(int value)
	{
		m_mark |= value;
	}

	public void nandMark(int value)
	{
		m_mark &= ~value;
	}

	public boolean isMarked(int value)
	{
		return((m_mark & value) != 0);
	}

	public boolean isSrcVisible()
	{
		return (m_src == drawSrc());
	}

	public boolean isDstVisible()
	{
		return (m_dst == drawDst());
	}

	/* Mark the edge as invalid so we paint it only once and then mark valid
	   -- not twice once from the source and once from the destination
	*/

	public void invalidateEdge()
	{
		nandMark(VALID_MARK);
	}

	public void setHighlightFlag() 
	{
		if (!isMarked(HIGHLIGHT_FLAG_MARK)) {
			orMark(HIGHLIGHT_FLAG_MARK);
			repaint();
	}	}

	public boolean getHighlightFlag() 
	{
		return isMarked(HIGHLIGHT_FLAG_MARK);
	}

	public void clearHighlightFlag() 
	{
		if (isMarked(HIGHLIGHT_FLAG_MARK)) {
			nandMark(HIGHLIGHT_FLAG_MARK);
			validate();
	}	}

	public void setGroupAndHighlightFlag() 
	{
		if ((m_mark & (GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK)) != (GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK)) {
			orMark(GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK);
			validate(); 
	}	}

	public void drawHighlighted() 
	{
		if (getHighlightFlag() && getRelationClass().isActive()) {
			draw(false);
	}	}

	public void clearGroupFlag() 
	{
		if (isMarked(GROUP_FLAG_MARK)) {
			nandMark(GROUP_FLAG_MARK);
			validate();
	}	}

	public boolean getGroupFlag() 
	{
		return isMarked(GROUP_FLAG_MARK);
	}

	// Compute the rules for how an edge is to be drawn between two entities, handling elision settings.
	// Essentially just flag the relation appropriately to assist the verify process know what it is dealing with

	public void draw(boolean allowElision) 
	{
		int				marks;
		boolean			src_elided, dst_elided;
		RelationClass	parentClass;

		if (isMarked(VALID_MARK)) {
			return;
		}

		parentClass     = getRelationClass();
		Diagram	diagram = getDiagram();


	/*	System.out.println("RelationInstance:draw(" + allowElision + ") " + this);
		java.lang.Thread.dumpStack();
		System.out.println("-----");
	*/
		marks = 0;
		if (!allowElision || parentClass.isClassVisible()) {
			if (m_src == null || m_dst == null) {
				MsgOut.println("Landscape error: null src or dst");
				return;
			}
					
			EntityInstance	src, dst, ancestor, drawRoot;

			src      = drawSrc();
			dst      = drawDst();

			if (src == null) {
				System.out.println("draw() " + this + " has no draw src");
				return;
			}
			if (dst == null) {
				System.out.println("draw() " + this + " has no draw dst");
				return;
			}

			drawRoot = diagram.getDrawRoot();
			if (drawRoot.hasDescendantOrSelf(src) || drawRoot.hasDescendantOrSelf(dst)) {
				// At least one edge addresses something in the diagram

				if (src != dst || src == m_src || dst == m_dst) {
					// Don't show internal edges beneath closed objects period
					// Such edges will have the same draw src and destination and both real edges lifted to this common closed ancestor

					marks = NORMAL_MARK;
					if (allowElision) {
						int				nid        = parentClass.getNid();
						EntityInstance	srcParent = src.getContainedBy();
						EntityInstance	dstParent = dst.getContainedBy();

						if (srcParent != null && srcParent == dstParent && srcParent.getElision(EntityInstance.INTERNAL_ELISION, nid)) {
							// Don't show -- this internal relation is elided
							marks = 0;
						} else {
							if (src.getElision(EntityInstance.SRC_ELISION, nid)) {
								marks |= OUT_ELIDED_MARK;
							} else {
								for (; srcParent != null && !srcParent.hasDescendantOrSelf(dst); srcParent = srcParent.getContainedBy()) {
									// Source goes from inside srcParent to outside of it
									if (srcParent.getElision(EntityInstance.EXITING_ELISION, nid)) {
										marks |= OUT_ELIDED_MARK;
										break;
							}	}	}

							if (dst.getElision(EntityInstance.DST_ELISION, nid)) {
								marks |= IN_ELIDED_MARK;
							} else {
								for (; dstParent != null && !dstParent.hasDescendantOrSelf(src); dstParent = dstParent.getContainedBy()) {
									// Destination is inside while src is outside
									if (dstParent.getElision(EntityInstance.ENTERING_ELISION, nid)) {
										marks |= IN_ELIDED_MARK;
										break;
					}	}	}	}	}

					if (src == dst) {
						if ((marks & NORMAL_MARK) != 0) {
							// Recursive loop
							marks |= LOOP_MARK;
						}
					} 
		}	}	}	
			
		JComponent relationComponent;

		if (marks == 0) {
			// Don't know how to draw this thing
			if (isMarked(IN_DIAGRAM_MARK)) {
				relationComponent = getSwingObject();
				if (relationComponent != null) {
					diagram.remove(relationComponent);
					setSwingObject(null);
				}
				nandMark(IN_DIAGRAM_MARK);
			}
		} else {
			marks |= VALID_MARK;
			if (allowElision) {
				marks |= ELISION_MARK;
			}
			relationComponent = neededComponent();

			if (!isMarked(IN_DIAGRAM_MARK)) {
				marks |= IN_DIAGRAM_MARK;
				orMark(marks);
				diagram.add(relationComponent);	// Add this relation into the diagram
				relationComponent.setVisible(true);
			} else {
				orMark(marks);
			}
			relationComponent.validate();
	}	}	

	public void writeRelation(PrintStream ps) throws IOException 
	{
		ps.print(qt(getParentClass().getId()) + " " + qt(m_src.getId()) + " " + qt(m_dst.getId()) + "\n");
	}

	public void writeAttributes(PrintStream ps) throws IOException 
	{
		ByteArrayOutputStream	byteStream = new ByteArrayOutputStream();
		PrintStream				ps1        = new PrintStream(byteStream);
		String					s;
		RelationClass			parentClass= getRelationClass();
		int						style      = getStyle();
		
		if (style >= 0) {
			if (parentClass == null || parentClass.getInheritedStyle() != style) {
				ps1.print(Attribute.indent + STYLE_ID + " = " + style + "\n");
		}	}

		super.writeAttributes(ps1, parentClass, false);

		s = byteStream.toString();
		if (s.length() != 0) {
			ps.print("(" + qt(parentClass.getId()) + " " + qt(m_src.getId()) + " " + qt(m_dst.getId()) + ") {\n" + s + "}\n");
		}
	}

	public EntityInstance getSrc() 
	{
		return m_src;
	}

	public void setSrc(EntityInstance src) 
	{
		m_src = src;
	}

	public EntityInstance getDst() 
	{
		return m_dst;
	}

	public EntityInstance getDrawSrc() 
	{
		return drawSrc();
	}

	public EntityInstance getCurrentSrc()
	{
		if (drawSrc() != null) {
			return(drawSrc());
		}
		return(m_src);
	}

	public EntityInstance getDrawDst() 
	{
		return drawDst();
	}
	
	public EntityInstance getCurrentDst()
	{
		if (drawDst() != null) {
			return(drawDst());
		}
		return(m_dst);
	}

	public RelationClass getRelationClass() 
	{
		return (RelationClass) getParentClass();
	}

	public boolean matches(RelationInstance other)
	{
		return(drawSrc() == other.drawSrc() && drawDst() == other.drawDst());
	}

	// Only called within an update

	protected void removeEdge()
	{
		getSrc().removeSrcRelation(this);
		getDst().removeDstRelation(this);
	}

	protected void deleteEdge()
	{
		Diagram diagram = getDiagram();
		removeEdge();
		if (diagram != null) {
			diagram.redrawDiagram();
	}	}

	class DeleteEdge extends MyUndoableEdit implements UndoableEdit
	{
		DeleteEdge()
		{
			logEdit(this);
		}

		public String getPresentationName() 
		{
			return "Delete " + RelationInstance.this.toString();
		}

		public void undo()
		{
			Diagram diagram = getDiagram();
			RelationInstance	ri = RelationInstance.this;
			
			ri.getSrc().addSrcRelation(ri);
			ri.getDst().addDstRelation(ri);
			diagram.redrawDiagram();

		}

		public void redo()
		{
			deleteEdge();
		}
	}	

	public void updateDeleteEdge()
	{
		deleteEdge();
		if (undoEnabled()) {
			new DeleteEdge();
	}	}

	public String toString() 
	{
		return drawSrc() + (drawSrc() == m_src ? "" : "{" + m_src + "}") + "->" + drawDst() + (drawDst() == m_dst ? "" : "{" + m_dst + "}");
	}

	// The routines that follow hide the complexity of getting/setting attribute values
	// from EditAttributes

	public int getPrimaryAttributeCount()
	{
		return(3);
	}

	public String getLsAttributeNameAt(int index)
	{
		String	name;

		switch (index) {
		case 0:
			name  = "class";
			break;
		case 1:
			name  = COLOR_ID;
			break;
		case 2:
			name  = STYLE_ID;
			break;	
		default:
			name  = super.getLsAttributeNameAt(index);
		}
		return(name);
	}

	public Object getLsAttributeValueAt(int index)
	{
		Object	value;

		switch (index) {
		case 0:
		{
			LandscapeClassObject parentClass = getParentClass();

			if (parentClass == null) {
				value = null;
			} else {
				value = parentClass.getLabelId();
			}
			break;
		}
		case 1:
			value = getObjectColor();
			break;
		case 2:
			value = new Integer(getStyle());
			break;
		default:
			value = super.getLsAttributeValueAt(index);
		}
		return(value);
	}

	public void updateAttributeValueAt(int index, Object value)
	{
		switch (index) {
		case 0:
			if (value != null) {
				LandscapeClassObject parentClass = getParentClass();
				String newId = (String) value;
				if (parentClass == null || !parentClass.getLabelId().equals(newId)) {
					Enumeration		en;
					RelationClass	ec;

					for (en = getTa().enumRelationClasses(); en.hasMoreElements(); ) {
						ec = (RelationClass) en.nextElement();
						if (ec.getLabelId().equals(newId)) {
							updateParentClass(ec);
							break;
			}	}	}	}
			break;
		case 1:
			updateObjectColor((Color) value);
			break;
		case 2:
			updateStyle(((Integer) value).intValue());
			break;
		default:
			super.updateAttributeValueAt(index, value);
	}	}

	// Need to know the type in cases where value might be null
	// For example with some colors

	public int getLsAttributeTypeAt(int index)
	{
		int		ret;
		
		switch (index) {
		case 0:
			ret = Attribute.RELATION_CLASS_TYPE;
			break;
		case 1:
			ret = Attribute.COLOR_OR_NULL_TYPE;
			break;
		case 2:
			ret = Attribute.REL_STYLE_TYPE;
			break;
		default:
			ret = super.getLsAttributeTypeAt(index);
		}
		return(ret);
	}

	public LandscapeEditorCore getLs()
	{
		return getDiagram().getLs();
	}

	public void mouseEntered()
	{
		getLs().setCursor(Cursor.HAND_CURSOR);
	}

	public void mouseExited()
	{
		getLs().setCursor(Cursor.DEFAULT_CURSOR); 
	}

	public void mousePressed(MouseEvent ev, int diagramX, int diagramY)
	{
		getDiagram().relationPressed(ev, this, diagramX, diagramY);
	}

	public void mouseReleased(MouseEvent ev, int diagramX, int diagramY)
	{
		getDiagram().relationReleased(ev, this, diagramX, diagramY);
	}

	public void mouseDragged(MouseEvent ev, int diagramX, int diagramY)
	{
		getDiagram().relationDragged(ev, this, diagramX, diagramY);
	}

	public void mouseMoved(MouseEvent ev, int diagramX, int diagramY)
	{
		Diagram diagram = getDiagram();

		if (this != g_infoShown) {
			LandscapeEditorCore ls = diagram.getLs();
			String				message;
			int					modifiers;


			modifiers = ev.getModifiers();
			if ((modifiers & Event.SHIFT_MASK) == 0) {
				ls.infoShown(this);
			}
			repaint();
			if (m_src == drawSrc()) {
				message = "";
			} else {
				message  = Util.quoted(m_src.getEntityLabel());
				if (isMarked(DIFFERENT_REAL_SRC_MARK)) {
					message += " et. al.";
				}
				message += "->";
			}
			message += Util.quoted(drawSrc().getEntityLabel()) +  " " + getClassLabel() + " " + Util.quoted(drawDst().getEntityLabel());
			if (m_dst != drawDst()) {
				message += "->" + Util.quoted(m_dst.getEntityLabel());
				if (isMarked(DIFFERENT_REAL_DST_MARK)) {
					message += " et. al.";
				}
			}

			if (m_freq != 1) {
				message += " (" + m_freq + " edges)";
			}
//			message += " " + (getX() + ev.getX()) + "x" + (getY() + ev.getY());
			
			getLs().showInfo(message);
		}
		diagram.movedOverThing(ev, this, diagramX, diagramY);
	}
}
